from marshmallow import Schema, class_registry
from abc import ABC, abstractmethod


class TiyaroBase(ABC):
    def __init__(self) -> None:
        self.inputSchema = None
        self.outputSchema = None

    def defInputSchema(self, schemaDict):
        self.inputSchema = Schema.from_dict(schemaDict, name="InputSchema")
        class_registry.register("InputSchema", self.inputSchema)

    def defOutputSchema(self, schemaDict):
        self.outputSchema = Schema.from_dict(schemaDict, name="OutputSchema")
        class_registry.register("OutputSchema", self.outputSchema)

    @abstractmethod
    def declareSchema(self):
        pass

    @abstractmethod
    def setup_model(self, pretrained_file_path):
        pass

    @abstractmethod
    def infer(self, json_input):
        pass


import base64
import hashlib
import json
import os

import click
import requests

from ..api.status_update import update_status_pushed
from ..handler.cli_state import get_model_name_with_suffix
from ..common.constants import *
from ..common.utils import (failed, get_model_endpoint, get_tiyaro_token,
                            success, warn)
from ..handler.cli_state import get_model_name_with_suffix
from ..handler.utils import validate_handler_exists


@click.command()
@click.option('-v', '--verbose', is_flag=True, default=False, help=VERBOSE_HELP)
def push(verbose):
    """
    - Pushes model repo to Tiyaro Infrastructure
    """
    do_push(verbose)


def push_model(tiyaro_jwt_token: str, is_verbose):
    tiyaro_model_name = get_model_name_with_suffix()
    tiyaro_model_version = MODEL_VERSION

    def tiyaro_get_name_reservation_record(api_token):
        headers = {
            "Authorization": f"Bearer {api_token}"
        }

        resp = requests.get(url=NAME_RESERVATION_ENDPOINT, headers=headers)
        if resp.status_code not in (200, 201):
            failed(
                f"Failed to reserve namespace.  Error code on {NAME_RESERVATION_ENDPOINT}: STATUS {resp.status}, REASON {resp.content}")
            warn('Kindly reach out to Tiyaro Support Team')
            exit(-1)
        return json.loads(resp.content)

    tiyaro_model_vendor = tiyaro_get_name_reservation_record(tiyaro_jwt_token)[
        'vendor_name']

    # TODO - USING PYTHON
    os.system(f"mkdir -p {UPLOAD_ARTIFACT_DIR}")
    # os.system("cp ./.tiyaro-install/state.json ./build") TODO: MULTI MODELS
    os.system(
        f"git archive --output={UPLOAD_ARTIFACT_DIR}/{GIT_SNAPSHOT_FILE_NAME} HEAD .")

    package_for_upload_client(UPLOAD_ARTIFACT_DIR, tiyaro_model_name,
                              tiyaro_model_vendor, tiyaro_model_version, tiyaro_jwt_token, is_verbose)

    success(
        f'\n***Uploaded {tiyaro_model_name} to be converted onto Tiyaro Infrastructure', is_verbose)
    return get_model_endpoint(tiyaro_model_vendor, tiyaro_model_version, tiyaro_model_name)


def package_for_upload_client(build_subdir, tiyaro_model_name, tiyaro_model_vendor, model_version, tiyaro_jwt_token, is_verbose):
    # TODO: Also send git hash info
    manifest_json = {
        'tiyaro_job_descriptor': TIYARO_JOB_DESCRIPTOR,
        'env_TIYARO_PUSH_NAME_BASE': None,
        'model_name': tiyaro_model_name,
        'model_vendor': tiyaro_model_vendor,
        'version': model_version,
        'model_type': TIYARO_MODEL_TYPE,
        'tiyaro_authorization_token': tiyaro_jwt_token
    }

    # Need to decorate with some tensor-specific information
    # with signature input and output

    manifest_json_filename = '/'.join([build_subdir, JSON_MANIFEST_FILE_NAME])
    with open(manifest_json_filename, 'w') as fout:
        fout.write(json.dumps(manifest_json, indent=4))

    required_header = 'file-upload-md5-binary-base64'

    success('START - Tarballing the artifacts into a single file', is_verbose)
    os.system(f'tar cf - {build_subdir} | gzip > {FILE_TO_UPLOAD}')
    success('FINISH - Tarballing the artifacts into a single file', is_verbose)

    success('START - Requesting a presigned url for upload', is_verbose)
    response_presigned_request = requests.get(
        PRESIGNED_REQUEST_ENDPOINT,
        headers={
            'Authorization': tiyaro_jwt_token
        })
    if response_presigned_request.status_code == 401:
        failed(
            f'Token Authorization Error.  Is your {TIYARO_TOKEN} still valid ?')
        exit(-1)
    elif response_presigned_request.status_code != 200:
        failed(
            f'Failed to received the presigned url from API endpoint {PRESIGNED_REQUEST_ENDPOINT}.', is_verbose)
        exit(-1)

    response_presigned_request_json = response_presigned_request.json()
    url = response_presigned_request_json['post_url']
    data = response_presigned_request_json['post_data_required']
    bucket_name = url.split('//')[1].split('.')[0]
    bucket_key = data['key']
    # click.echo(f'INFO - Received a presigned for S3 bucket {bucket_name} and key {bucket_key}')
    success(f'FINISH - Received a presigned url for upload {url}', is_verbose)

    size_payload = os.path.getsize(FILE_TO_UPLOAD)
    success(
        f'START - Upload of payload size {str(round(size_payload / (1024 * 1024), 2))}MB to a presigned url')
    with open(FILE_TO_UPLOAD, 'rb') as test_file_to_upload:
        # the key supposed to be file may be
        files = {'file': test_file_to_upload}
        response_file_upload = requests.post(url, data=data, files=files)
    if response_file_upload.status_code != 204:
        failed(
            f'Failed to upload the {FILE_TO_UPLOAD} to presigned url')
        exit(-1)

    success(f'FINISH - Upload to a presigned url')
    os.remove(FILE_TO_UPLOAD)
    return bucket_name, bucket_key


def do_push(is_verbose):
    validate_handler_exists()
    token = get_tiyaro_token()
    api = push_model(token, is_verbose)
    update_status_pushed(get_model_name_with_suffix())

    success('\n\nView status at:')
    success(STATUS_PAGE_URL)
    success('\nInference API:')
    success(api)



import os

import click

from .constants import INFERENCE_ENDPOINT_PREFIX, TIYARO_TOKEN


def wip():
    warn(f'** WORK IN PROGRESS **\n Please wait for new version release..')


def get_tiyaro_token():
    token = os.getenv(TIYARO_TOKEN, None)
    if token is None:
        failed(f'Please set {TIYARO_TOKEN} env var')
        exit(-1)
    return token


def get_model_endpoint(vendor, version, model):
    return f'{INFERENCE_ENDPOINT_PREFIX}/{vendor}/{version}/{model}'


def success(msg, is_verbose=True):
    if is_verbose:
        click.secho(msg, fg='green')


def failed(msg, is_verbose=True):
    if is_verbose:
        click.secho(msg, fg='red')


def warn(msg, is_verbose=True):
    if is_verbose:
        click.secho(msg, fg='yellow')



import imp
import os

import yaml

from ..common.constants import HANDLER_MODEL_MANIFEST_FILE, MODEL_VERSION
from .utils import validate_model_manifest_file_exists

#TODO: reintroduce when we add support for version
# def get_model_version():
#     validate_model_manifest_file_exists()
#     return f'{MODEL_VERSION}'

# # use this only for getting value from manifest file
# def get_model_name():
#     validate_model_manifest_file_exists()
#     return _get_param('name')

def _get_param(field):
    with open(HANDLER_MODEL_MANIFEST_FILE, 'r') as file:
        contents = yaml.safe_load(file)
        value = contents[field]
        if (
            (value is None)
            or (not isinstance(value, str))
            or (not value.strip())
        ):
            click.echo(
                f'Please set a valid model {field} in {HANDLER_MODEL_MANIFEST_FILE}')
        return value



import os
import shutil
import subprocess
import tarfile

import click
import requests

from ..api.status_update import update_status_init
from ..common.constants import *
from ..common.utils import failed, get_tiyaro_token, success, warn
from ..handler.cli_state import get_model_name_with_suffix, save_model_name


@click.command()
@click.option('-f', '--force', is_flag=True, default=False, help=INIT_HELP)
@click.option('-v', '--verbose', is_flag=True, default=False, help=VERBOSE_HELP)
def init(force, verbose):
    """
    - Initializes model repo for Tiyaro Push
    """
    name = click.prompt('Please enter name of your model', type=str)
    get_templates(force, verbose)
    save_model_name(name, verbose)
    update_status_init(get_model_name_with_suffix())
    success(
        f'Created {TIYARO_HANDLER_DIR}, {HANDLER_MODEL_MODEL_TEST_FILE} templates successfully !')


def get_templates(is_overwrite, is_verbose):

    if (os.path.isdir(TIYARO_HANDLER_DIR)):
        if not is_overwrite:
            warn(
                f'{TIYARO_HANDLER_DIR} already exists.  To force init, kindly use -f with your init command')
            exit(-1)
        else:
            if is_verbose:
                subprocess.run(f'tiyaro clear -f -v', shell=True)
            else:
                subprocess.run(f'tiyaro clear -f', shell=True)

    if (os.path.isfile(HANDLER_MODEL_MODEL_TEST_FILE)):
        if not is_overwrite:
            warn(f'{HANDLER_MODEL_MODEL_TEST_FILE} already exists.  To force init kindly use -f with your init command')
            exit(-1)
        else:
            subprocess.run(f'tiyaro clear -f', shell=True)

    token = get_tiyaro_token()
    resp = requests.get(
        f'{PUSH_SUPPORT_FILES_ENDPOINT}/{ARTIFACTS_FILE_NAME}',
        headers={
            'Authorization': token
        })
    if resp.status_code == 200:
        template_url = resp.content
    else:
        failed(resp.status_code)
        failed(resp.content)
        failed(
            f'Unable to get templates URL.  Is your {TIYARO_TOKEN} still valid ?')
        exit(-1)

    os.makedirs(ARTIFACTS_DOWNLOAD_DIR)
    downloaded_artifact = f'{ARTIFACTS_DOWNLOAD_DIR}/ARTIFACTS_FILE_NAME'

    resp = requests.get(template_url, stream=True)
    if resp.status_code == 200:
        with open(downloaded_artifact, 'wb') as f:
            f.write(resp.raw.read())
    else:
        failed(
            f'Unable to get templates.  Is your {TIYARO_TOKEN} still valid ?')
        exit(-1)

    def members(tf, sub_folder):
        l = len(sub_folder)
        for member in tf.getmembers():
            if member.path.startswith(sub_folder):
                member.path = member.path[l:]
                yield member

    tar = tarfile.open(downloaded_artifact)
    tar.extractall(path=TIYARO_HANDLER_DIR,
                   members=members(tar, ARTIFACTS_FILES_DIR))
    # move test file to project root for tiyaro test
    shutil.move(f'{TIYARO_HANDLER_DIR}/{HANDLER_MODEL_MODEL_TEST_FILE}',
                f'{HANDLER_MODEL_MODEL_TEST_FILE}')
    shutil.rmtree(ARTIFACTS_DOWNLOAD_DIR)



import subprocess
import venv
import os
import shutil

import click

from ..api.status_update import (update_status_test_failed,
                                 update_status_test_passed)
from ..handler.cli_state import get_model_name_with_suffix
from ..common.constants import (HANDLER_MODEL_MANIFEST_FILE,
                                HANDLER_MODEL_MODEL_TEST_FILE, VERBOSE_HELP, TIYARO_TEST_VENV_PATH)
from ..common.utils import failed, success
from ..handler.utils import (validate_handler_exists,
                             validate_handler_test_file_exists)


@click.command()
@click.option('-p', '--pretrained', required=False, help=f'pretrained_file path or url.  default is value from {HANDLER_MODEL_MANIFEST_FILE}')
@click.option('-i', '--input', required=True, help=f'Valid JSON input string or file path')
@click.option('-v', '--verbose', is_flag=True, default=False, help=VERBOSE_HELP)
def test(pretrained, input, verbose):
    """
    - Test model locally
    """
    do_test(pretrained, input, verbose)


def do_test(pretrained, input, is_verbose):
    validate_handler_exists()
    validate_handler_test_file_exists()
    success(f'Creating VENV: {TIYARO_TEST_VENV_PATH}', is_verbose)
    venv.create(TIYARO_TEST_VENV_PATH, with_pip=True)
    # TODO: get requirements file from handler manifest
    os.system(f'{TIYARO_TEST_VENV_PATH}/bin/pip install -r requirements.txt')

    if not pretrained:
        pretrained = 'from-config'

    p = subprocess.run(
        f'{TIYARO_TEST_VENV_PATH}/bin/python {HANDLER_MODEL_MODEL_TEST_FILE} {pretrained} {input}', shell=True)
    if p.returncode == 0:
        success('Test successful !  You can push your model now.')
        if (os.path.isdir(TIYARO_TEST_VENV_PATH)):
            success(f'Cleaning up VENV: {TIYARO_TEST_VENV_PATH}', is_verbose)
            shutil.rmtree(TIYARO_TEST_VENV_PATH)
            update_status_test_passed(get_model_name_with_suffix())
    else:
        failed('Test failed !')
        update_status_test_failed(get_model_name_with_suffix())
